/*
 * Copyright (c) 1986 Eric P. Allman
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "sendmail.h"

#ifndef lint
#ifdef NAMED_BIND
static char sccsid[] = "@(#)domain.c	5.22 (Berkeley) 6/1/90 (with name server)";
#else
static char sccsid[] = "@(#)domain.c	5.22 (Berkeley) 6/1/90 (without name server)";
#endif
#endif /* not lint */

#ifdef NAMED_BIND

#include <sys/param.h>
#include <errno.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <netdb.h>

typedef union {
	HEADER qb1;
	char qb2[PACKETSZ];
} querybuf;

static char hostbuf[MAXMXHOSTS*PACKETSZ];

getmxrr(host, mxhosts, rcode)
	char *host, **mxhosts;
	int *rcode;
{
	extern int h_errno;
	register u_char *eom, *cp;
	register int i, j, n, nmx;
	register char *bp;
	HEADER *hp;
	querybuf answer;
	int ancount, qdcount, buflen, seenlocal, options;
	u_short pref, localpref, type, prefer[MAXMXHOSTS];

	errno = 0;
	options = _res.options;
	_res.options &= ~RES_DNSRCH;
	n = res_search(host, C_IN, T_MX, (char *)&answer, sizeof(answer));
	_res.options = options;
	if (n < 0)
	{
		if (tTd(8, 1))
			printf("getmxrr: res_search failed (errno=%d, h_errno=%d)\n",
			    errno, h_errno);
		switch (h_errno)
		{
		  case NO_DATA:
		  case NO_RECOVERY:
			/* no MX data on this host */
			goto punt;

		  case HOST_NOT_FOUND:
			/* the host just doesn't exist */
			*rcode = EX_NOHOST;
			break;

		  case TRY_AGAIN:
			/* couldn't connect to the name server */
			if (!UseNameServer && errno == ECONNREFUSED)
				goto punt;

			/* it might come up later; better queue it up */
			*rcode = EX_TEMPFAIL;
			break;
		}

		/* irreconcilable differences */
		return (-1);
	}

	/* find first satisfactory answer */
	hp = (HEADER *)&answer;
	cp = (u_char *)&answer + sizeof(HEADER);
	eom = (u_char *)&answer + n;
	for (qdcount = ntohs(hp->qdcount); qdcount--; cp += n + QFIXEDSZ)
		if ((n = dn_skipname(cp, eom)) < 0)
			goto punt;
	nmx = 0;
	seenlocal = 0;
	localpref = 0;
	buflen = sizeof(hostbuf);
	bp = hostbuf;
	ancount = ntohs(hp->ancount);
	while (--ancount >= 0 && cp < eom && nmx < MAXMXHOSTS) {
		if ((n = dn_expand((char *)&answer, eom, cp, bp, buflen)) < 0)
			break;
		cp += n;
		GETSHORT(type, cp);
 		cp += sizeof(u_short) + sizeof(u_long);
		GETSHORT(n, cp);
		if (type != T_MX)  {
			if (tTd(8, 1) || _res.options & RES_DEBUG)
				printf("unexpected answer type %d, size %d\n",
				    type, n);
			cp += n;
			continue;
		}
		GETSHORT(pref, cp);
		if ((n = dn_expand((char *)&answer, eom, cp, bp, buflen)) < 0)
			break;
		cp += n;
		if (host_is_local(bp)) {
			if (seenlocal == 0 || pref < localpref)
				localpref = pref;
			seenlocal = 1;
			continue;
		}
		prefer[nmx] = pref;
		mxhosts[nmx++] = bp;
		n = strlen(bp) + 1;
		bp += n;
		buflen -= n;
	}
	if (nmx == 0) {
punt:		mxhosts[0] = strcpy(hostbuf, host);
		return(1);
	}

	/* sort the records */
	for (i = 0; i < nmx; i++) {
		for (j = i + 1; j < nmx; j++) {
			if (prefer[i] > prefer[j] ||
			    (prefer[i] == prefer[j] && rand() % 1 == 0)) {
				register int temp;
				register char *temp1;

				temp = prefer[i];
				prefer[i] = prefer[j];
				prefer[j] = temp;
				temp1 = mxhosts[i];
				mxhosts[i] = mxhosts[j];
				mxhosts[j] = temp1;
			}
		}
		if (seenlocal && prefer[i] >= localpref) {
			/*
			 * truncate higher pref part of list; if we're
			 * the best choice left, we should have realized
			 * awhile ago that this was a local delivery.
			 */
			if (i == 0) {
				*rcode = EX_CONFIG;
				return(-1);
			}
			nmx = i;
			break;
		}
	}
	return(nmx);
}

int
host_is_local(h)
	char *h;
{
	STAB *s = stab(h, ST_CLASS, ST_FIND);
	return (s != NULL && bitnset('w', s->s_class));
}

#ifdef DECWRL

getcanonname(host, hbsize)
	char *host;
	int hbsize;
{
	extern int h_errno;
	register u_char *eom, *cp;
	register int n; 
	HEADER *hp;
	querybuf answer;
	u_short type;
	int ancount, qdcount, options;
	char nbuf[PACKETSZ];

	/* we don't have to worry about CNAME's here since they must
	 * ultimately point to an MX or A and the resolver will take
	 * care of resolving them for us so we just ask for MX and A.
	 *
	 * if you ask for an MX or A and there's a CNAME, you get the
	 * CNAME followed by an MX or A.  so what we do is scan the
	 * response looking for the MX or A.  we do all this with
	 * DNSRCH turned off since partial matches on mail hosts are
	 * unsecure as well as tremendously expensive in DNS bandwidth.
	 *
	 * we don't pay a lot of attention to search order, since
	 * this routine is only called to expand a name to its
	 * canonical form (or in IDA/KJS, to decide whether the name
	 * is worth an SMTP delivery attempt or not).
	 */
	
	options = _res.options;
	_res.options &= ~RES_DNSRCH;
	n = res_search(host, C_IN, T_MX,
			(char *)&answer, sizeof(answer));
	if (tTd(8, 1))
		printf("getcanonname: MX=%d\n", n);

	if (n < 0) {
		n = res_search(host, C_IN, T_A,
				(char *)&answer, sizeof(answer));
		if (tTd(8, 1))
			printf("getcanonname:  no MX, but A=%d\n", n);
	}
	_res.options = options;
	if (n < 0) {
		if (tTd(8, 1))
			printf("getcanonname:  res_search failed (errno=%d, h_errno=%d)\n",
			    errno, h_errno);
		return FALSE;
	}

	/* find first satisfactory answer */
	hp = (HEADER *)&answer;
	ancount = ntohs(hp->ancount);

	/* we don't care about errors here, only if we got an answer */
	if (ancount == 0) {
		if (tTd(8, 1))
			printf("rcode = %d, ancount=%d\n", hp->rcode, ancount);
		return FALSE;
	}
	/* init our walkers and then skip the question section */
	cp = (u_char *)&answer + sizeof(HEADER);
	eom = (u_char *)&answer + n;
	for (qdcount = ntohs(hp->qdcount); qdcount--; cp += n + QFIXEDSZ)
		if ((n = dn_skipname(cp, eom)) < 0)
			return FALSE;

	for (;  ancount > 0 && cp < eom;  ancount--, cp += n) {
		n = dn_expand((char *)&answer, eom, cp, nbuf, sizeof(nbuf));
		if (n < 0)
			break;
		cp += n;
		GETSHORT(type, cp);
 		cp += sizeof(u_short) + sizeof(u_long);
		GETSHORT(n, cp);
		if (type == T_MX || type == T_A) {
			/* note that nbuf is the question, not the answer.
			 * so on MX, the name we expand is the one we asked
			 * for (perhaps after passing through a CNAME), not
			 * the MX-name.
			 */
			(void)strncpy(host, nbuf, hbsize);
			host[hbsize - 1] = '\0';
			return TRUE;
		}
	}
	/* if there were no MX or A RR's in the response, the resolver
	 * is insane.  we deal with it as we would deal with a pink
	 * elephant in our living room: that is, pretend it isn't there.
	 */
	return FALSE;
}

#else DECWRL

getcanonname(host, hbsize)
	char *host;
	int hbsize;
{
	struct hostent *hp;

	hp = gethostbyname(host);
	if (hp == NULL)
		return FALSE;

	if (strlen(hp->h_name) >= hbsize)
		return FALSE;

	(void) strcpy(host, hp->h_name);
	return TRUE;
}

#endif DECWRL

#endif NAMED_BIND
